"""
requests events module.
Currently it instruments only botocore.vendored.requests.
For regular requests lib, we use urllib3
"""

from __future__ import absolute_import
import traceback
import json
from uuid import uuid4

from epsagon.utils import add_data_if_needed
from ..trace import trace_factory
from ..event import BaseEvent
from ..http_filters import is_blacklisted_url
from ..utils import update_http_headers, normalize_http_url
from ..constants import HTTP_ERR_CODE


class RequestsEvent(BaseEvent):
    """
    Represents base requests event.
    """

    ORIGIN = 'requests'
    RESOURCE_TYPE = 'http'

    # pylint: disable=W0613
    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """
        super(RequestsEvent, self).__init__(start_time)

        self.event_id = 'requests-{}'.format(str(uuid4()))

        prepared_request = args[0]
        self.resource['name'] = normalize_http_url(prepared_request.url)
        self.resource['operation'] = prepared_request.method
        self.resource['metadata']['url'] = prepared_request.url

        add_data_if_needed(
            self.resource['metadata'],
            'request_headers',
            dict(prepared_request.headers)
        )

        epsagon_trace_id = prepared_request.headers.get('epsagon-trace-id')
        # Make sure trace ID is present in case headers will be removed.
        if epsagon_trace_id:
            self.resource['metadata']['http_trace_id'] = epsagon_trace_id

        add_data_if_needed(
            self.resource['metadata'],
            'request_body',
            prepared_request.body
        )

        if response is not None:
            self.update_response(response)

        if exception is not None:
            self.set_exception(exception, traceback.format_exc())

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        :return: None
        """

        self.resource['metadata']['status_code'] = response.status_code
        self.resource = update_http_headers(
            self.resource,
            response.headers
        )

        add_data_if_needed(
            self.resource['metadata'],
            'response_headers',
            dict(response.headers)
        )

        # Extract only json responses
        self.resource['metadata']['response_body'] = None
        try:
            add_data_if_needed(
                self.resource['metadata'],
                'response_body',
                json.loads(response.content)
            )
        except ValueError:
            pass

        # Detect errors based on status code
        if response.status_code >= HTTP_ERR_CODE:
            self.set_error()


class RequestsEventFactory(object):
    """
    Factory class, generates requests event.
    """

    @staticmethod
    def create_event(wrapped, instance, args, kwargs, start_time, response,
                     exception):
        """
        Create an event according to the given api_name.
        """
        prepared_request = args[0]
        # Detect if URL is blacklisted, and ignore.
        if is_blacklisted_url(prepared_request.url):
            return

        event = RequestsEvent(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        trace_factory.add_event(event)



"""
pymysql patcher module
"""
from __future__ import absolute_import

import wrapt
from .db_wrapper import connect_wrapper


def patch():
    """
    patch module.
    :return: None
    """
    wrapt.wrap_function_wrapper(
        'pymysql',
        'connect',
        connect_wrapper
    )



"""
Utilities for Epsagon module.
"""

from __future__ import absolute_import, print_function
import os
import collections
import socket
import sys
import traceback
import re
import six
import requests
import simplejson as json
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

from epsagon import http_filters
from epsagon.constants import TRACE_COLLECTOR_URL, REGION
from .trace import trace_factory, create_transport
from .constants import EPSAGON_HANDLER, DEBUG_MODE, DEFAULT_SAMPLE_RATE


METADATA_CACHE = {
    'queried': False,
    'data': {},
}


def normalize_http_url(url):
    """
    Strip http schema, port number and path from a url
    :param url: the url to normalize
    :return: normalized url
    """
    parsed = urlparse(url)
    netloc = parsed.netloc
    return netloc.split(':')[0] if netloc else url


def add_data_if_needed(dictionary, name, data):
    """
    Add data to the given dictionary if metadata_only option is set to False.
    :param dictionary: dictionary to add the data to
    :param name: key name
    :param data: value
    :return: None
    """
    dictionary[name] = None
    if not trace_factory.metadata_only:
        dictionary[name] = data


def update_http_headers(resource_data, response_headers):
    """
    Updates resource data dict with AWS entities if matching header found.
    :param resource_data: event's resource data dict
    :param response_headers: response headers from HTTP request
    :return: update resource data dict
    """
    for header_key, header_value in response_headers.items():
        if header_key.lower() == 'x-amzn-requestid':
            # This is a request to API Gateway
            if '.appsync-api.' not in resource_data['metadata']['url']:
                resource_data['type'] = 'api_gateway'
            resource_data['metadata']['request_trace_id'] = header_value
            break

    return resource_data


def get_tc_url(use_ssl):
    """
    Get the TraceCollector URL.
    :return: TraceCollector URL.
    """
    protocol = 'https://' if use_ssl else 'http://'

    return TRACE_COLLECTOR_URL.format(protocol=protocol, region=REGION)


def init(
    token='',
    app_name='Application',
    collector_url=None,
    metadata_only=True,
    disable_timeout_send=False,
    use_ssl=True,
    debug=False,
    send_trace_only_on_error=False,
    url_patterns_to_ignore=None,
    keys_to_ignore=None,
    keys_to_allow=None,
    ignored_endpoints=None,
    split_on_send=False,
    propagate_lambda_id=False,
    logging_tracing_enabled=True,
    step_dict_output_path=None,
    sample_rate=DEFAULT_SAMPLE_RATE,
):
    """
    Initializes trace with user's data.
    User can configure here trace parameters.
    :param token: user's token
    :param app_name: application name
    :param collector_url: the url of the collector.
    :param metadata_only: whether to send only the metadata, or also the data.
    :param disable_timeout_send: whether to disable traces send on timeout
     (when enabled, is t done using a signal handler).
    :param use_ssl: whether to use SSL or not.
    :param debug: debug mode flag,
    :param send_trace_only_on_error: Whether to send trace only when
     there is an error or not.
    :param url_patterns_to_ignore: URL patterns to ignore in HTTP data
      collection.
    :param keys_to_ignore: List of keys to ignore while extracting metadata.
    :param keys_to_allow: List of keys to allow while extracting metadata
    :param ignored_endpoints: List of ignored endpoints for web frameworks.
    :param split_on_send: Split the trace on send flag
    :param propagate_lambda_id: Inject identifiers via return value flag
    :param logging_tracing_enabled: Add an epsagon log id to logging calls
    :param step_dict_output_path:
        Path in the result dict to append the Epsagon steps data
    :param sample_rate: A number between 0 and 1, represents the probability
        of a trace to be sent.
        When enabled (value < 1), sampling will be performed according to
        the given value.
    :return: None
    """
    if not collector_url:
        collector_url = get_tc_url(
            ((os.getenv('EPSAGON_SSL') or '').upper() == 'TRUE') | use_ssl
        )
    # Ignored URLs is a comma separated values, if coming from env.
    ignored_urls = os.getenv('EPSAGON_URLS_TO_IGNORE')
    if ignored_urls:
        ignored_urls = ignored_urls.split(',')

    # Ignored URLs is a comma separated values, if coming from env.
    ignored_paths = os.getenv('EPSAGON_ENDPOINTS_TO_IGNORE')
    if ignored_paths:
        ignored_paths = ignored_paths.split(',')

    # Same goes for Ignored keys.
    ignored_keys = os.getenv('EPSAGON_IGNORED_KEYS')
    if ignored_keys:
        ignored_keys = ignored_keys.split(',')

    allowed_keys = os.getenv('EPSAGON_ALLOWED_KEYS')
    if allowed_keys:
        allowed_keys = allowed_keys.split(',')

    step_dict_output_path_env = os.getenv('EPSAGON_STEPS_OUTPUT_PATH')
    if step_dict_output_path_env:
        step_dict_output_path_env = step_dict_output_path_env.split('.')

    # If EPSAGON_METADATA exists as an env var - use it
    if os.getenv('EPSAGON_METADATA'):
        metadata_only = (os.getenv('EPSAGON_METADATA') or '').upper() == 'TRUE'

    # If EPSAGON_LOGGING_TRACING_ENABLED exists as an env var - use it
    if os.getenv('EPSAGON_LOGGING_TRACING_ENABLED'):
        logging_tracing_enabled = (
            os.getenv('EPSAGON_LOGGING_TRACING_ENABLED') or ''
        ).upper() == 'TRUE'

    # In case we're running on AWS Lambda, logging correlation is disabled
    if is_lambda_env():
        logging_tracing_enabled = False

    if os.getenv('EPSAGON_SAMPLE_RATE'):
        sample_rate = float(os.getenv('EPSAGON_SAMPLE_RATE'))

    trace_factory.initialize(
        token=os.getenv('EPSAGON_TOKEN') or token,
        app_name=os.getenv('EPSAGON_APP_NAME') or app_name,
        collector_url=os.getenv('EPSAGON_COLLECTOR_URL') or collector_url,
        metadata_only=metadata_only,
        disable_timeout_send=(
            ((os.getenv('EPSAGON_DISABLE_ON_TIMEOUT') or '')
                .upper() == 'TRUE')
            | disable_timeout_send
        ),
        debug=((os.getenv('EPSAGON_DEBUG') or '')
               .upper() == 'TRUE') | debug,
        send_trace_only_on_error=(
            ((os.getenv('EPSAGON_SEND_TRACE_ON_ERROR') or '')
                .upper() == 'TRUE')
            | send_trace_only_on_error
        ),
        url_patterns_to_ignore=ignored_urls or url_patterns_to_ignore,
        keys_to_ignore=ignored_keys or keys_to_ignore,
        keys_to_allow=allowed_keys or keys_to_allow,
        transport=create_transport(collector_url, token),
        split_on_send=(
                ((os.getenv('EPSAGON_SPLIT_ON_SEND') or '').upper() == 'TRUE')
                | split_on_send
        ),
        propagate_lambda_id=(
                ((os.getenv('EPSAGON_PROPAGATE_LAMBDA_ID') or '').upper() ==
                 'TRUE')
                | propagate_lambda_id
        ),
        logging_tracing_enabled=logging_tracing_enabled,
        step_dict_output_path=(
            step_dict_output_path_env or step_dict_output_path
        ),
        sample_rate=sample_rate
    )

    # Append to ignored endpoints
    http_filters.add_ignored_endpoints(ignored_paths or ignored_endpoints)


def import_original_module():
    """
    Imports original user's handler using the `EPSAGON_HANDLER` env.
    :return: Module or Exception
    """
    original_handler = os.getenv(EPSAGON_HANDLER)
    if not original_handler:
        raise ValueError(
            'EPSAGON_HANDLER value not specified in environment variable'
        )

    try:
        module_path, handler_name = original_handler.rsplit('.', 1)
    except ValueError:
        raise ValueError('wrong handler formatted: {}'.format(original_handler))

    module_path = module_path.replace('/', '.')

    try:
        return __import__(module_path), module_path, handler_name
    except ImportError:
        raise ImportError('Failed to import module: {}'.format(module_path))


def collect_container_metadata(metadata):
    """
    Collects container metadata if exists.
    :return: dict.
    """
    is_k8s = os.environ.get('KUBERNETES_SERVICE_HOST')
    if is_k8s:
        metadata['is_k8s'] = True
        metadata['k8s_pod_name'] = socket.gethostname()
        with open('/proc/self/cgroup', 'r') as proc_file:
            metadata['k8s_container_id'] = (
                proc_file.readline().split('/')[-1].rstrip('\n')
            )

    if METADATA_CACHE['queried'] and METADATA_CACHE['data']:
        metadata['ECS'] = METADATA_CACHE['data']
        return

    METADATA_CACHE['queried'] = True
    metadata_uri = os.environ.get('ECS_CONTAINER_METADATA_URI')
    if not metadata_uri:
        return
    container_metadata = json.loads(requests.get(metadata_uri).content)

    new_metadata = container_metadata['Labels'].copy()
    new_metadata['Limits'] = container_metadata['Limits']
    METADATA_CACHE['data'] = new_metadata
    if METADATA_CACHE['data']:
        metadata['ECS'] = METADATA_CACHE['data']


def find_in_object(obj, key, path=None):
    """
    recursively search for a key in an object
    :param obj: The dict to search in
    :param key: The key to search for
    :param path: The path (as an array) of the key nested in the object
    :return: The value and path of this key in the object,
        or None if was not found
    """
    if not path:
        path = []

    if isinstance(obj, collections.Mapping):
        # search key in obj
        if key in obj:
            return obj[key], path

    if (
        isinstance(obj, collections.Iterable)
        and not isinstance(obj, six.string_types)
    ):
        for k in obj:
            # Handle lists as well
            element = k if isinstance(obj, list) else obj[k]
            nested_path = None if isinstance(obj, list) else path + [k]
            result = find_in_object(element, key, nested_path)
            if result:
                return result

    return None

def collect_exception_python3(exception):
    """
    Collect exception from exception __traceback__.
    :param exception: Exception from Flask.
    :return: traceback data
    """

    traceback_data = ''.join(traceback.format_exception(
        type(exception),
        exception,
        exception.__traceback__,
    ))
    return traceback_data


def collect_exception_python2():
    """
    Collect exception from exception sys.exc_info.
    :return: traceback data
    """

    traceback_data = six.StringIO()
    traceback.print_exception(*sys.exc_info(), file=traceback_data)
    return traceback_data.getvalue()


def get_traceback_data_from_exception(exception):
    """
    Get traceback data from exception
    :param exception: the Exception
    :return: traceback data
    """
    python_version = sys.version_info.major
    if python_version == 2:
        return collect_exception_python2()
    if python_version == 3:
        return collect_exception_python3(exception)
    return ''


def parse_json(json_string):
    """
    Parse JSON string to a Python Dictionary
    :param json_string: JSON string
    :return: Python Dictionary
    """
    try:
        return json.loads(json_string)
    except ValueError:
        return None


def camel_case_to_title_case(camel_case_string):
    """
    Turn Camel Case string into Title Case string in which first characters of
    all the words are capitalized.
    :param camel_case_string: Camel Case string
    :return: Title Case string
    """
    if not isinstance(camel_case_string, str):
        return None
    title_case = re.sub('([^-])([A-Z][a-z-]+)', r'\1 \2', camel_case_string)\
        .title()
    return title_case


def add_metadata_from_dict(resource, dictionary, key):
    """
    Add new data to resource metadata
    :param resource: Resource
    :param dictionary: Argument Dictionary
    :param key: Property key
    :return: True if added, else False
    """
    value = dictionary.get(key)
    if not value and not isinstance(value, str):
        return
    title_case_key = camel_case_to_title_case(key)
    resource['metadata'][title_case_key] = value


def is_lambda_env():
    """
    Returns True if the current environment is running on a Lambda function.
    :return: bool
    """
    return os.getenv('AWS_LAMBDA_FUNCTION_NAME') is not None


def print_debug(log):
    if DEBUG_MODE:
        print('[EPSAGON_DEBUG]: {}'.format(log))



"""General constants"""

import os

__version__ = '1.51.0'

DEFAULT_REGION = 'us-east-1'
REGION = os.getenv('AWS_REGION', DEFAULT_REGION)

TRACE_COLLECTOR_URL = '{protocol}{region}.tc.epsagon.com'
COLD_START = True

DEBUG_MODE = ((os.getenv('EPSAGON_DEBUG') or '').upper() == 'TRUE')

# Customer original handler.
EPSAGON_HANDLER = 'EPSAGON_HANDLER'

TIMEOUT_GRACE_TIME_MS = 200
# How long we try to send traces in seconds.
TIMEOUT_ENV = float(os.getenv('EPSAGON_SEND_TIMEOUT_SEC', '0'))
SEND_TIMEOUT = TIMEOUT_ENV if TIMEOUT_ENV else TIMEOUT_GRACE_TIME_MS / 1000.0

MAX_LABEL_SIZE = 10 * 1024

DEFAULT_SAMPLE_RATE = 1

# User-defined HTTP minimum status code to be treated as an error.
HTTP_ERR_CODE = int(os.getenv('EPSAGON_HTTP_ERR_CODE', '500'))

# List of ignored endpoints for web frameworks.
IGNORED_ENDPOINTS = []

EPSAGON_MARKER = '__EPSAGON'

STRONG_KEYS = [
    'key',
    'request_id',
    'requestid',
    'request-id',
    'steps_dict',
    'message_id',
    'etag',
    'item_hash',
    'sequence_number',
    'trace_id',
    'job_id',
    'activation_id',
    'hostname',
    'virtual_host',
]


def is_strong_key(key):
    """
    Checks if given key is a strong key
    :param key: key
    :return: is a strong key
    """
    key = key.replace(' ', '_').lower()
    for strong_key in STRONG_KEYS:
        if strong_key in key:
            return True
    return False


STEP_DICT_NAME = 'Epsagon'
EPSAGON_EVENT_ID_KEY = '_epsagon_event_id'
TRACE_URL_PREFIX = (
    'https://app.epsagon.com/functions/{aws_account}/{region}/{function_name}'
    '?requestId={request_id}&requestTime={request_time}'
)



"""
pymongo patcher module.
"""

from __future__ import absolute_import
import wrapt
from epsagon.modules.general_wrapper import wrapper
from ..events.pymongo import PyMongoEventFactory


def _wrapper(wrapped, instance, args, kwargs):
    """
    General wrapper for pymongo instrumentation.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """

    return wrapper(PyMongoEventFactory, wrapped, instance, args, kwargs)


def patch():
    """
    Patch module.
    :return: None
    """

    wrapt.wrap_function_wrapper(
        'pymongo.collection',
        'Collection.insert_one',
        _wrapper
    )
    wrapt.wrap_function_wrapper(
        'pymongo.collection',
        'Collection.insert_many',
        _wrapper
    )



"""
General wrapper for instrumentation.
"""

#pylint: disable=W0703
from __future__ import absolute_import
import time
import traceback
from epsagon.trace import trace_factory


def wrapper(factory, wrapped, instance, args, kwargs):
    """
    General wrapper for instrumentation.
    :param factory: Factory class for the event type
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """

    response = None
    exception = None
    start_time = time.time()

    try:
        response = wrapped(*args, **kwargs)
        return response
    except Exception as operation_exception:
        exception = operation_exception
        raise
    finally:
        try:
            factory.create_event(
                wrapped,
                instance,
                args,
                kwargs,
                start_time,
                response,
                exception
            )
        except Exception as instrumentation_exception:
            trace_factory.add_exception(
                instrumentation_exception,
                traceback.format_exc()
            )



"""
sqlalchemy events module.
"""

from __future__ import absolute_import
from uuid import uuid4
import traceback

try:
    from psycopg2.extensions import parse_dsn

except ImportError:
    def parse_dsn(dsn):
        """
        Parse the DSN.
        :param dsn: input DSN.
        :return:
        """
        return dict(
            attribute.split('=') for attribute in dsn.split()
            if '=' in attribute
        )

from ..trace import trace_factory
from ..event import BaseEvent

MAX_QUERY_SIZE = 2048


class DBAPIEvent(BaseEvent):
    """
    Represents base sqlalchemy event.
    """

    ORIGIN = 'dbapi'
    RESOURCE_TYPE = 'database'
    RESOURCE_OPERATION = None

    # mapping SQL commands to words preceding the table name in the query
    # Note: Not supporting advanced syntax of select and delete (as 'delete
    # from only ..')
    _OPERATION_TO_TABLE_NAME_KEYWORD = {
        'select': 'from',
        'insert': 'into',
        'update': 'update',
        'delete': 'from',
        'create': 'table'
    }

    def __init__(
            self,
            connection,
            cursor,
            _args,
            _kwargs,
            start_time,
            exception
    ):
        """
        Initialize.
        :param connection: The SQL engine the event is using
        :param cursor: Cursor object used in the even
        :param args: args passed to called function
        :param kwargs: kwargs passed to called function
        :param start_time: Start timestamp (epoch)
        :param exception: Exception (if occurred)
        """

        super(DBAPIEvent, self).__init__(start_time)

        self.event_id = 'dbapi-{}'.format(str(uuid4()))

        # in case of pg instrumentation we extract data from the dsn property
        if hasattr(connection, 'dsn'):
            dsn = parse_dsn(connection.dsn)
            db_name = dsn['dbname']
            host = dsn.get('host', 'local')
            query = cursor.query
        else:
            query = _args[0]
            host = connection.extract_hostname
            db_name = connection.extract_dbname

        self.resource['name'] = db_name if db_name else host

        # NOTE: The operation might not be identified properly when
        # using 'WITH' clause
        operation = query.split()[0].lower()
        self.resource['operation'] = operation

        # override event type with the specific DB type
        if 'rds.amazonaws' in host:
            self.resource['type'] = 'rds'
        elif 'redshift.amazonaws' in host:
            self.resource['type'] = 'redshift'

        self.resource['metadata'] = {
            'Host': host,
            'Driver': connection.__class__.__module__.split('.')[0],
            'Table Name': self._extract_table_name(query, operation)
        }

        # for select we always want to save the query
        if (
                (operation == 'select') or
                (not trace_factory.metadata_only)
        ):
            self.resource['metadata']['Query'] = query[:MAX_QUERY_SIZE]

        if exception is None:
            # Update response data
            self.resource['metadata']['Related Rows Count'] = int(
                cursor.rowcount
            )
        else:
            self.set_exception(exception, traceback.format_exc())

    @staticmethod
    def _extract_table_name(query, operation):
        """
        Extract the table name from the SQL query string
        :param query: The SQL query string
        :param operation: The SQL operation used in the query
            (SELECT, INSERT, etc.)
        :return: Table name (string), "" if couldn't find
        """

        if operation in DBAPIEvent._OPERATION_TO_TABLE_NAME_KEYWORD:
            keyword = DBAPIEvent._OPERATION_TO_TABLE_NAME_KEYWORD[operation]
            query_words = query.lower().split()
            if keyword in query_words:
                return query.split()[query_words.index(keyword) + 1]

        return ''


class DBAPIEventFactory(object):
    """
    Factory class, generates dbapi event.
    """

    @staticmethod
    # pylint: disable=W0613
    def create_event(wrapped, cursor_wrapper, args, kwargs, start_time,
                     response, exception):
        """
        Create an event according to the given operation.
        :param wrapped:
        :param cursor_wrapper:
        :param args:
        :param kwargs:
        :param start_time:
        :param response:
        :param exception:
        :return:
        """
        event = DBAPIEvent(
            cursor_wrapper.connection_wrapper,
            cursor_wrapper,
            args,
            kwargs,
            start_time,
            exception,
        )
        trace_factory.add_event(event)



"""
Runner for Azure Functions.
"""

from __future__ import absolute_import
import os
from ..event import BaseEvent


class AzureFunctionRunner(BaseEvent):
    """
    Represents Azure function event runner.
    """

    ORIGIN = 'runner'
    RESOURCE_TYPE = 'azure_function'
    OPERATION = 'Invoke'

    def __init__(self, start_time, context):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param context: function's context, azure.functions.Context
        """

        super(AzureFunctionRunner, self).__init__(start_time)

        self.event_id = context.invocation_id
        self.resource['name'] = context.function_name
        self.resource['operation'] = self.OPERATION

        self.resource['metadata'].update({
            'azure.resource_group': os.getenv('ResourceGroupName', ''),
            'azure.location': os.getenv('Location', ''),
            'azure.function.log': os.getenv('LOGNAME', ''),
            'azure.function.app': os.getenv('WEBSITE_SITE_NAME', ''),
        })



"""
Runner for a Flask Python function
"""

from __future__ import absolute_import
import uuid
from ..event import BaseEvent
from ..utils import add_data_if_needed, normalize_http_url


class FlaskRunner(BaseEvent):
    """
    Represents Python Flask event runner.
    """

    ORIGIN = 'runner'
    RESOURCE_TYPE = 'python_flask'
    OPERATION = 'request'

    def __init__(self, start_time, app, request):
        """
        Initialize.
        :param start_time: event's start time (epoch).
        :param app: the flask application.
        :param request: the incoming request.
        """

        super(FlaskRunner, self).__init__(start_time)

        self.event_id = str(uuid.uuid4())

        self.resource['name'] = (
            normalize_http_url(request.headers.get('Host'))
            if request.headers and request.headers.get('Host')
            else app.name
        )
        self.resource['operation'] = request.method

        self.resource['metadata'].update({
            'Base URL': request.base_url,
            'Path': request.path,
            'User Agent': request.headers.get('User-Agent', 'N/A'),
            'Flask Application': app.name,
            'Endpoint': request.endpoint,
        })

        if request.query_string:
            self.resource['metadata']['Query String'] = request.query_string

        if request.data:
            add_data_if_needed(
                self.resource['metadata'],
                'Request Data',
                request.data
            )

        add_data_if_needed(
            self.resource['metadata'],
            'Request Headers',
            dict(request.headers)
        )

        if request.values:
            add_data_if_needed(
                self.resource['metadata'],
                'Request Values',
                dict(request.values)
            )

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: WSGI Response
        :return: None
        """

        add_data_if_needed(
            self.resource['metadata'],
            'Response Data',
            response.data
        )

        add_data_if_needed(
            self.resource['metadata'],
            'Response Headers',
            dict(response.headers)
        )

        self.resource['metadata']['Status'] = response.status

        if response.status_code >= 500:
            self.set_error()



"""
httplib2 events module.
"""

from __future__ import absolute_import
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse
import traceback
from uuid import uuid4
import simplejson as json

from epsagon.utils import add_data_if_needed
from ..trace import trace_factory
from ..event import BaseEvent
from ..http_filters import (
    is_blacklisted_url,
    is_payload_collection_blacklisted
)
from ..utils import update_http_headers
from ..constants import HTTP_ERR_CODE


class Httplib2Event(BaseEvent):
    """
    Represents base gttplib2 event.
    """

    ORIGIN = 'httplib2'
    RESOURCE_TYPE = 'http'

    #pylint: disable=W0613
    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """

        super(Httplib2Event, self).__init__(start_time)
        self.event_id = 'httplib2-{}'.format(str(uuid4()))

        # Params can be set via args or kwargs.
        url, method, body, headers = Httplib2Event.unroller(*args, **kwargs)

        url_obj = urlparse(url)
        self.resource['name'] = url_obj.hostname
        self.resource['operation'] = method
        self.resource['metadata']['url'] = url

        if not is_payload_collection_blacklisted(url):
            if headers:
                add_data_if_needed(
                    self.resource['metadata'],
                    'request_headers',
                    headers
                )

            try:
                if body:
                    add_data_if_needed(
                        self.resource['metadata'],
                        'request_body',
                        json.loads(body)
                    )
            except (TypeError, json.errors.JSONDecodeError):
                # Skip if it is not a JSON body
                pass

        if response is not None:
            self.update_response(response)

        if exception is not None:
            self.set_exception(exception, traceback.format_exc())

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        :return: None
        """

        response_headers, response_body = response

        self.resource['metadata']['status'] = int(response_headers['status'])
        self.resource = update_http_headers(
            self.resource,
            response_headers
        )

        full_url = self.resource['metadata']['url']

        if not is_payload_collection_blacklisted(full_url):
            add_data_if_needed(
                self.resource['metadata'],
                'response_headers',
                dict(response_headers)
            )

            # Extract only json responses
            try:
                if response_body:
                    add_data_if_needed(
                        self.resource['metadata'],
                        'response_body',
                        json.loads(response_body)
                    )
            except (TypeError, json.errors.JSONDecodeError):
                # Skip if it is not a JSON body
                pass

        # Detect errors based on status code
        if int(response_headers['status']) >= HTTP_ERR_CODE:
            self.set_error()

    @staticmethod
    def unroller(uri='N/A', method='N/A', body=None, headers=None):
        return uri, method, body, headers


class Httplib2EventFactory(object):
    """
    Factory class, generates Httplib2 event.
    """

    @staticmethod
    def create_event(wrapped, instance, args, kwargs, start_time, response,
                     exception):
        """
        Create an event according to the given api_name.
        """
        url, _, _, _ = Httplib2Event.unroller(*args, **kwargs)

        # Detect if URL is blacklisted, and ignore.
        if is_blacklisted_url(url):
            return

        event = Httplib2Event(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        trace_factory.add_event(event)



"""
flask patcher module.
"""

from __future__ import absolute_import
import wrapt
from ..wrappers.flask import FlaskWrapper
from ..utils import print_debug, is_lambda_env


def _wrapper(wrapped, instance, args, kwargs):
    """
    Adds `FlaskWrapper` into Flask app.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    """
    # Skip on Lambda environment since it's not relevant and might be duplicate
    if is_lambda_env():
        return wrapped(*args, **kwargs)

    response = wrapped(*args, **kwargs)
    try:
        FlaskWrapper(instance)
    except Exception:  # pylint: disable=broad-except
        print_debug('Could not add Flask wrapper')
    return response


def patch():
    """
    Patch module.
    :return: None
    """
    wrapt.wrap_function_wrapper('flask', 'Flask.__init__', _wrapper)



"""
Azure sdk events module.
"""

# pylint: disable=C0302
from __future__ import absolute_import

import traceback
from uuid import uuid4
from importlib import import_module
from ..trace import trace_factory
from ..event import BaseEvent
from ..utils import add_data_if_needed

# Conditionally importing Azure error class
AzureError = Exception  # pylint: disable=invalid-name
try:
    AzureError = (  # pylint: disable=invalid-name
        import_module('azure.core.exceptions').AzureError
    )
except ImportError:
    pass


# pylint: disable=W0613
def empty_func(*args):
    """
    A dummy function.
    :return: None
    """


class AzureEvent(BaseEvent):
    """
    Represents base Azure SDK event.
    """

    ORIGIN = 'azure-sdk'
    RESOURCE_TYPE = 'azure-sdk'
    RESPONSE_TO_FUNC = {}
    OPERATION_TO_FUNC = {}

    # pylint: disable=W0613
    def __init__(
            self,
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
    ):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """
        super(AzureEvent, self).__init__(start_time)
        self.event_id = 'azure-{}'.format(str(uuid4()))
        self.resource['operation'] = wrapped.__name__
        self.resource['metadata'] = {}

        if response is not None:
            self.update_response(response)

        if exception is not None:
            self.set_exception(exception, traceback.format_exc())

    def update_response(self, _response):
        """
        Adds response data to event.
        :param _response: Response from azure
        :return: None
        """
        self.RESPONSE_TO_FUNC.get(self.resource['operation'], empty_func)()

    def set_exception(self, exception, traceback_data, handled=True):
        """
        see {Event.set_exception}
        """
        super(AzureEvent, self).set_exception(
            exception,
            traceback_data,
            handled=handled
        )

        # Specific handling for azure errors
        if isinstance(exception, AzureError):
            self.event_id = exception.headers.get('x-ms-activity-id')
            self.resource['metadata']['azure_error'] = True
            self.resource['metadata']['status_code'] = exception.status_code
            self.resource['metadata']['error_message'] = exception.message
            self.resource['metadata']['reason'] = exception.reason


class AzureCosmosContainerEvent(AzureEvent):
    """
    Represents CosmosDB Container Azure event.
    """

    RESOURCE_TYPE = 'cosmos_db'

    def __init__(
            self,
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
    ):
        """
        Initialize.
        """
        super(AzureCosmosContainerEvent, self).__init__(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )
        self.RESPONSE_TO_FUNC.update({
            'upsert_item': self.upsert_item_response,
        })

        self.OPERATION_TO_FUNC.update({
            'upsert_item': self.upsert_item_request,
            'delete_item': self.delete_item_request,
            'query_items': self.query_items_request,
        })
        self.kwargs = kwargs
        self.args = args
        self.response = response

        self.resource['metadata']['azure.location'] = (
            instance.client_connection.ReadEndpoint.split(
                '-'
            )[1].split('.')[0]
        )
        self.resource['metadata']['azure.cosmos.endpoint'] = (
            instance.client_connection.url_connection
        )
        _, database, _, container = instance.container_link.split('/')
        self.resource['name'] = container
        self.resource['metadata']['azure.cosmos.database_id'] = database
        self.resource['metadata']['azure.cosmos.container_id'] = container

        self.OPERATION_TO_FUNC.get(self.resource['operation'], empty_func)()

    def upsert_item_request(self):
        self.resource['metadata']['azure.cosmos.document_id'] = (
            self.args[0].get('id')
        )

    def query_items_request(self):
        self.resource['metadata']['azure.cosmos.query'] = (
            self.kwargs.get('query')
        )

    def delete_item_request(self):
        self.resource['metadata']['azure.cosmos.document_id'] = (
            self.args[0].get('id')
        )
        add_data_if_needed(
            self.resource['metadata'],
            'azure.cosmos.document',
            self.args[0]
        )

    def upsert_item_response(self):
        self.resource['metadata']['azure.cosmos.document.etag'] = (
            self.kwargs.get('_etag')
        )
        add_data_if_needed(
            self.resource['metadata'],
            'azure.cosmos.document',
            self.response
        )


class AzureEventFactory(object):
    """
    Factory class, generates azure sdk event.
    """

    CLASS_MAPPING = {
        'ContainerProxy': AzureCosmosContainerEvent,
    }

    @staticmethod
    def create_event(
        wrapped,
        instance,
        args,
        kwargs,
        start_time,
        response,
        exception
    ):
        """
        Create an event according to the given instance_type.
        """

        event_class = AzureEventFactory.CLASS_MAPPING.get(
            instance.__class__.__name__
        )

        if event_class is not None:
            event = event_class(
                wrapped,
                instance,
                args,
                kwargs,
                start_time,
                response,
                exception
            )
            trace_factory.add_event(event)



"""
Wrapper for a general python function
"""

from __future__ import absolute_import
import time
import traceback
import functools
import epsagon.trace
import epsagon.runners.python_function
from epsagon.utils import collect_container_metadata
from epsagon import constants


def wrap_python_function(func, args, kwargs, name=None):
    """
    Wrap a python function call with a simple python wrapper. Used as default
    when wrapping with other wrappers is impossible.
    NOTE: this function does not prepare the tracer (clears the previous run)
    :param func: The function to wrap.
    :param args: The arguments to the function.
    :param kwargs: The keyword arguments to the function.
    :param name: Resource name for the runner
    :return: The function's result.
    """
    try:
        runner = epsagon.runners.python_function.PythonRunner(
            time.time(),
            func,
            args,
            kwargs,
            name=name
        )
        epsagon.trace.trace_factory.set_runner(runner)

        # Collect metadata in case this is a container.
        collect_container_metadata(runner.resource['metadata'])

    # pylint: disable=W0703
    except Exception:
        # If we failed, just call the user's function. Nothing more to do.
        return func(*args, **kwargs)

    # settings in case we are in a lambda and context is None
    constants.COLD_START = False

    result = None
    try:
        result = func(*args, **kwargs)
        return result
    # pylint: disable=W0703
    except Exception as exception:
        runner.set_exception(exception, traceback.format_exc())
        raise
    finally:
        try:
            runner.add_json_field('python.function.return_value', result)
        # pylint: disable=W0703
        except Exception as exception:
            epsagon.trace.trace_factory.add_exception(
                exception,
                traceback.format_exc(),
            )
        try:
            epsagon.trace.trace_factory.send_traces()
        # pylint: disable=W0703
        except Exception:
            pass


def python_wrapper(*args, **kwargs):
    """
    Epsagon's general Python wrapper.

    Receives optional keyword arg 'name': used to identify this resource in
        the application (defaults to function name).

    Options for using:
    -   @python_wrapper(name='my-resource')
        def my_function(params):
            ...

    -   @python_wrapper()
        def my_function(params):
            ...

    -   @python_wrapper
        def my_function(params):
            ...

    NOTE: Should not be used for two functions in the same stack trace, or when
          running under a traced context (e.g. AWS Lambda, HTTP frameworks, etc)
    """
    name = kwargs.get('name')

    def _inner_wrapper(func):

        @functools.wraps(func)
        def _python_wrapper(*args, **kwargs):
            epsagon.trace.trace_factory.get_or_create_trace().prepare()
            return wrap_python_function(func, args, kwargs, name=name)

        return _python_wrapper

    if len(args) == 1 and callable(args[0]):
        return _inner_wrapper(args[0])

    return _inner_wrapper



# pylint: disable=protected-access
"""
Runner for a Tornado Python framework
"""

from __future__ import absolute_import
import uuid
from ..event import BaseEvent
from ..utils import add_data_if_needed

MAX_PAYLOAD_BYTES = 2000


class TornadoRunner(BaseEvent):
    """
    Represents Python Tornado event runner.
    """

    ORIGIN = 'runner'
    RESOURCE_TYPE = 'python_tornado'

    def __init__(self, start_time, request):
        """
        Initialize.
        :param start_time: event's start time (epoch).
        :param request: the incoming request.
        """

        super(TornadoRunner, self).__init__(start_time)

        self.event_id = str(uuid.uuid4())

        # Since Tornado doesn't has app name, we use the tracer app name.
        self.resource['name'] = request.host
        self.resource['operation'] = request.method

        self.resource['metadata'].update({
            'Host': request.host,
            'Protocol': request.protocol,
            'Path': request.path,
            'Version': request.version,
            'Remote IP': request.remote_ip,
            'User Agent': request.headers.get('User-Agent', 'N/A'),
        })

        if request.query:
            add_data_if_needed(
                self.resource['metadata'],
                'Query',
                request.query
            )

    def update_response(self, response, response_body=None):
        """
        Adds response data to event.
        :param response_body: Response body
        :param response: WSGI Response
        """
        headers = dict(response._headers.get_all())
        add_data_if_needed(
            self.resource['metadata'],
            'Response Headers',
            headers
        )

        if response_body:
            add_data_if_needed(
                self.resource['metadata'],
                'Response Body',
                str(response_body)[:MAX_PAYLOAD_BYTES]
            )

        self.resource['metadata']['Status'] = response._status_code
        self.resource['metadata']['etag'] = headers.get('Etag')

        if not self.error_code and response._status_code >= 500:
            self.set_error()



"""trace transport layers"""

import base64
import logging
import requests
import simplejson as json
from epsagon.constants import SEND_TIMEOUT
from epsagon.trace_encoder import TraceEncoder


def to_json(obj):
    return json.dumps(obj, cls=TraceEncoder, encoding='latin1')


class NoneTransport(object):
    @classmethod
    def send(cls, _):
        logging.error('trace sent using NoneTransport, configure a transport')


class LogTransport(object):
    """ send traces by logging them """

    @staticmethod
    def send(trace):
        trace_json = to_json(trace.to_dict())
        trace_message = base64.b64encode(
            trace_json.encode('utf-8')
        ).decode('utf-8')

        # pylint: disable=superfluous-parens
        print('EPSAGON_TRACE: {}'.format(trace_message))


class HTTPTransport(object):
    """ send traces using http request """

    def __init__(self, dest, token):
        self.dest = dest
        self.token = token
        self.timeout = SEND_TIMEOUT
        self.session = requests.Session()

    def send(self, trace):
        headers = {'Authorization': 'Bearer {}'.format(self.token)}
        self.session.post(self.dest,
                          data=to_json(trace.to_dict()),
                          headers=headers,
                          timeout=self.timeout)



"""
google events module.
"""

from __future__ import absolute_import
from uuid import uuid4
import traceback

from epsagon.utils import add_data_if_needed
from ..event import BaseEvent
from ..trace import trace_factory


class GoogleRPCEvent(BaseEvent):
    """
    Represents base google rpc event
    """

    ORIGIN = 'grpc'
    RESOURCE_TYPE = 'grpc'

    #pylint: disable=W0613
    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """

        super(GoogleRPCEvent, self).__init__(start_time)

        request_data = args[0]

        _, endpoint, operation = getattr(instance, '_method').split('/')
        endpoint = endpoint.split('.')[-1]

        self.event_id = 'grpc-{}'.format(str(uuid4()))
        self.resource['operation'] = operation
        self.resource['name'] = endpoint

        add_data_if_needed(
            self.resource['metadata'],
            'Request Data',
            str(request_data)
        )

        if response is not None:
            self.update_response(response)

        if exception is not None:
            self.set_exception(exception, traceback.format_exc())

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        :return: None
        """

        add_data_if_needed(
            self.resource['metadata'],
            'Response Data',
            str(response)
        )


class GRPCNaturalLanguageEvent(GoogleRPCEvent):
    """
    Represents Natural Language grpc event.
    """

    RESOURCE_TYPE = 'LanguageService'


class GRPCEventFactory(object):
    """
    Factory class, generates google rpc event.
    """

    FACTORY = {
        class_obj.RESOURCE_TYPE: class_obj
        for class_obj in GoogleRPCEvent.__subclasses__()
    }

    @staticmethod
    def create_event(wrapped, instance, args, kwargs, start_time, response,
                     exception):
        """
        Create an event according to the given endpoint.
        :param wrapped:
        :param instance:
        :param args:
        :param kwargs:
        :param start_time:
        :param response:
        :param exception:
        :return:
        """
        _, endpoint, _ = getattr(instance, '_method').split('/')
        endpoint = endpoint.split('.')[-1]

        event_class = GRPCEventFactory.FACTORY.get(endpoint, GoogleRPCEvent)
        event = event_class(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )
        trace_factory.add_event(event)



"""
Triggers for Azure Function
"""

from __future__ import absolute_import
from uuid import uuid4
from epsagon.utils import add_data_if_needed
from ..event import BaseEvent
try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse


class BaseAzureTrigger(BaseEvent):
    """
    Represents base Azure Function trigger
    """
    ORIGIN = 'trigger'


class HTTPAzureTrigger(BaseAzureTrigger):
    """
    Represents an HTTP Azure Function trigger.
    """
    RESOURCE_TYPE = 'http'

    def __init__(self, start_time, event, response):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point,
            azure.functions.HttpRequest
        :param response: http response, azure.functions.HttpResponse
        """

        super(HTTPAzureTrigger, self).__init__(start_time)
        self.resource['operation'] = event.method
        url_data = urlparse(event.url)
        self.resource['name'] = url_data.netloc

        self.event_id = event.headers.get('x-arr-log-id', str(uuid4()))

        self.resource['metadata'] = {
            'http.request.path': url_data.path,
            'http.status_code': response.status_code,
        }

        if event.params:
            add_data_if_needed(
                self.resource['metadata'],
                'http.request.path_params',
                dict(event.params.items())
            )

        add_data_if_needed(
            self.resource['metadata'],
            'http.request.headers',
            event.headers.__http_headers__
        )

        add_data_if_needed(
            self.resource['metadata'],
            'http.response.headers',
            response.headers.__http_headers__
        )

        try:
            add_data_if_needed(
                self.resource['metadata'],
                'http.request.body',
                event.get_json()
            )
        except Exception:  # pylint: disable=broad-except
            pass


class AzureTriggerFactory(object):
    """
    Represents a Azure Function trigger Factory.
    """

    @staticmethod
    def factory(start_time, event, result):
        """
        Creates trigger event object.
        :param start_time: event's start time (epoch)
        :param event: event dict from the entry point
        :param result: function result, can be relevant in HTTP
        :return: Event or None
        """
        if 'req' in event:
            return HTTPAzureTrigger(start_time, event['req'], result)
        return None



"""
Common tests helpers
"""
import mock


def get_tracer_patch_kwargs():
    return {
        'metadata_only': False,
        'disable_timeout_send': False,
        'prepare': mock.MagicMock(),
        'send_traces': mock.MagicMock(),
        'events': [],
        'add_event': mock.MagicMock(),
        'add_exception': mock.MagicMock(),
        'set_runner': mock.MagicMock()
    }


import epsagon.wrappers.python_function
from epsagon.trace import trace_factory
import epsagon.runners.python_function
import epsagon.constants
import mock
import requests

TEST_URL = 'https://test.com/'

def setup_function(func):
    trace_factory.use_single_trace = True
    epsagon.constants.COLD_START = True

@mock.patch('urllib3.poolmanager.PoolManager.connection_from_url')
def test_pool_manager_patching(pool_manager_mock):
    retval = 'success'

    @epsagon.wrappers.python_function.python_wrapper
    def wrapped_function():
        requests.get(TEST_URL)
        return retval

    assert wrapped_function() == retval
    pool_manager_mock.assert_called_with(TEST_URL)



import epsagon.wrappers.python_function
from epsagon.trace import trace_factory
import epsagon.runners.python_function
import epsagon.constants
import logging


def setup_function(func):
    trace_factory.use_single_trace = True
    epsagon.constants.COLD_START = True


def test_logging_exception_capture(trace_transport):
    retval = 'success'

    @epsagon.wrappers.python_function.python_wrapper
    def wrapped_function(event, context):
        logging.exception('test')
        return retval

    assert wrapped_function('a', 'b') == retval

    exception = trace_transport.last_trace.events[0].exception
    assert exception['type'] == 'EpsagonException'
    assert exception['message'] == 'test'


def test_logging_exception_capture_with_args(trace_transport):
    retval = 'success'

    @epsagon.wrappers.python_function.python_wrapper
    def wrapped_function(event, context):
        logging.exception('test %s %s', 'test', 'test')
        return retval

    assert wrapped_function('a', 'b') == retval

    exception = trace_transport.last_trace.events[0].exception
    assert exception['type'] == 'EpsagonException'
    assert exception['message'] == 'test test test'



"""
urllib3 patcher module.
"""

from __future__ import absolute_import
import uuid
import wrapt
from epsagon.modules.general_wrapper import wrapper
from ..events.urllib3 import Urllib3EventFactory
from ..http_filters import is_blacklisted_url


def _get_headers_from_args(
        method=None,
        url=None,
        body=None,
        headers=None,
        retries=None,
        redirect=None,
        assert_same_host=True,
        timeout=None,
        pool_timeout=None,
        release_conn=None,
        chunked=False,
        body_pos=None,
        **response_kw
):
    """
    extract headers from arguments
    """
    # pylint: disable=unused-argument
    # not using '_' in arg names so unrolling will be smoother
    return headers


def _wrapper(wrapped, instance, args, kwargs):
    """
    General wrapper for requests instrumentation.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """
    # Inject header to support tracing over HTTP requests to
    # opentracing monitored code
    trace_id = uuid.uuid4().hex
    span_id = uuid.uuid4().hex[16:]
    parent_span_id = uuid.uuid4().hex[16:]

    host_url = '{}://{}'.format(instance.scheme, instance.host)

    # Detect if URL is blacklisted, and ignore.
    if not is_blacklisted_url(host_url):
        headers = _get_headers_from_args(*args, **kwargs)
        if headers is None:  # explicitly checking None to not catch {}
            if len(args) >= 4:
                # we got None headers as in args[3]
                args = list(args)
                headers = args[3] = {}
                args = tuple(args)
            else:
                # either kwargs['headers'] == None or it doesn't exist
                headers = kwargs['headers'] = {}

        headers['epsagon-trace-id'] = (
            '{trace_id}:{span_id}:{parent_span_id}:1'.format(
                trace_id=trace_id,
                span_id=span_id,
                parent_span_id=parent_span_id
            )
        )

    return wrapper(Urllib3EventFactory, wrapped, instance, args, kwargs)


def patch():
    """
    Patch module.
    :return: None
    """

    try:
        wrapt.wrap_function_wrapper(
            'urllib3',
            'HTTPConnectionPool.urlopen',
            _wrapper
        )
    except Exception:  # pylint: disable=broad-except
        # Can happen in different Python versions.
        pass



import mock
import pytest
from epsagon import trace_factory
import epsagon.constants

trace_mock = mock.MagicMock()


def setup_function(func):
    trace_factory.use_single_trace = True


def test_function_wrapper_sanity(trace_transport, ):
    retval = 'success'

    @epsagon.python_wrapper(name='test-func')
    def wrapped_function(event, context):
        return retval

    assert wrapped_function('a', 'b') == 'success'

    assert len(trace_transport.last_trace.events) == 1

    event = trace_transport.last_trace.events[0]
    assert event.resource['type'] == 'python_function'
    assert event.resource['name'] == 'test-func'
    assert event.resource['metadata']['python.function.return_value'] == retval
    assert event.error_code == 0


def test_function_wrapper_function_exception(trace_transport):
    @epsagon.python_wrapper()
    def wrapped_function(event, context):
        raise TypeError('test')

    with pytest.raises(TypeError):
        wrapped_function('a', 'b')

    assert len(trace_transport.last_trace.events) == 1

    event = trace_transport.last_trace.events[0]
    assert event.exception['type'] == 'TypeError'
    assert event.resource['metadata']['python.function.return_value'] is None
    assert not trace_transport.last_trace.exceptions


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
def test_python_wrapper_python_runner_factory_failed(_):
    @epsagon.python_wrapper
    def wrapped_function(event, context):
        return 'success'

    with mock.patch(
            'epsagon.runners.python_function.PythonRunner',
            side_effect=TypeError()
    ):
        assert wrapped_function('a', 'b') == 'success'

    trace_mock.prepare.assert_called_once()
    trace_mock.send_traces.assert_not_called()
    trace_mock.set_runner.assert_not_called()



"""
requests patcher module.
"""

from __future__ import absolute_import
import wrapt
from epsagon.modules.general_wrapper import wrapper
from ..events.requests import RequestsEventFactory
from ..constants import EPSAGON_MARKER
from ..utils import print_debug


def _wrapper(wrapped, instance, args, kwargs):
    """
    General wrapper for requests instrumentation.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """
    # Marking connection pool so requests won't be captured in urllib3 as well
    try:
        for adapter in instance.adapters.values():
            connection_pool = adapter.poolmanager.connection_from_url(
                args[0].url
            )
            setattr(connection_pool, EPSAGON_MARKER, True)
    except Exception as exception:  # pylint: disable=broad-except
        print_debug('Could not add marker to requests adapter: {}'.format(
            exception
        ))
    return wrapper(RequestsEventFactory, wrapped, instance, args, kwargs)


def patch():
    """
    Patch module.
    :return: None
    """

    wrapt.wrap_function_wrapper(
        'requests',
        'Session.send',
        _wrapper
    )



"""
Wrappers module.
"""
from __future__ import absolute_import

from .aws_lambda import lambda_wrapper, step_lambda_wrapper
from .chalice import chalice_wrapper
from .azure_function import azure_wrapper
from .python_function import python_wrapper
from .gcp_function import gcp_wrapper

__all__ = [
    'lambda_wrapper',
    'azure_wrapper',
    'python_wrapper',
    'step_lambda_wrapper',
    'gcp_wrapper',
    'chalice_wrapper',
]



"""
Celery patcher module.
This uses the built-in Signals in Celery to get signal for every event
"""

from __future__ import absolute_import
from importlib import import_module
from ..events.celery import (
    wrap_prerun,
    wrap_postrun,
    wrap_before_publish,
    wrap_after_publish,
    wrap_retry,
    wrap_failure,
)


def patch():
    """
    Patch module.
    :return: None
    """
    signals = import_module('celery.signals')
    signals.before_task_publish.connect(wrap_before_publish, weak=False)
    signals.after_task_publish.connect(wrap_after_publish, weak=False)
    signals.task_prerun.connect(wrap_prerun, weak=False)
    signals.task_retry.connect(wrap_retry, weak=False)
    signals.task_failure.connect(wrap_failure, weak=False)
    signals.task_postrun.connect(wrap_postrun, weak=False)



"""
Runner for AWS Lambda.
"""

from __future__ import absolute_import
import os
from uuid import uuid4
from ..event import BaseEvent
from .. import constants
from ..common import ErrorCode


class AbstractLambdaRunner(BaseEvent):
    """
    Represents Lambda event runner.
    """
    ORIGIN = 'runner'
    RESOURCE_TYPE = NotImplemented
    OPERATION = 'invoke'
    ARN_WITH_ALIAS_LENGTH = 8
    AWS_ACCOUNT_IND = 4

    def __init__(self, start_time, context):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        :param context: Lambda's context (passed from entry point)
        """

        super(AbstractLambdaRunner, self).__init__(start_time)

        # Creating a unique ID for local runs.
        self.event_id = (
            context.aws_request_id
            if context.aws_request_id != '1234567890'
            else 'local-{}'.format(str(uuid4()))
        )
        self.resource['name'] = context.function_name
        self.resource['operation'] = self.OPERATION

        arn_split = context.invoked_function_arn.split(':')
        self.resource['metadata'].update({
            'log_stream_name': context.log_stream_name,
            'log_group_name': context.log_group_name,
            'function_version': context.function_version,
            'memory': context.memory_limit_in_mb,
            'aws_account': arn_split[self.AWS_ACCOUNT_IND],
            'cold_start': constants.COLD_START,
            'region': os.getenv('AWS_REGION', ''),
        })

        # Extract Function alias if exists
        if len(arn_split) == self.ARN_WITH_ALIAS_LENGTH:
            self.resource['metadata']['function_alias'] = arn_split[-1]

    def set_timeout(self):
        """
        Sets timeout error code.
        :return: None
        """
        # Don't override exceptions
        if self.error_code != ErrorCode.EXCEPTION:
            self.error_code = ErrorCode.TIMEOUT


class LambdaRunner(AbstractLambdaRunner):
    """
    Represents Lambda event runner.
    """
    RESOURCE_TYPE = 'lambda'


class StepLambdaRunner(AbstractLambdaRunner):
    """
    Represents Lambda event runner.
    """
    RESOURCE_TYPE = 'step_function_lambda'

    def add_step_data(self, steps_dict):
        """
        Add steps function data.
        :param steps_dict: The steps dictionary to add.
        """
        self.resource['metadata']['steps_dict'] = steps_dict



"""
botocore patcher module.
"""

from __future__ import absolute_import
import json
from uuid import uuid4
import wrapt
from epsagon.modules.general_wrapper import wrapper
from epsagon.constants import STEP_DICT_NAME
from ..events.botocore import (
    BotocoreEventFactory,
    BotocoreStepFunctionEvent
)
from .requests import _wrapper as _requests_wrapper


def _wrapper(wrapped, instance, args, kwargs):
    """
    General wrapper for botocore instrumentation.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """
    instance_type = instance.__class__.__name__.lower()
    if instance_type == BotocoreStepFunctionEvent.RESOURCE_TYPE:
        handle_stepfunc_args(args)

    return wrapper(BotocoreEventFactory, wrapped, instance, args, kwargs)


def add_steps_dict_to_request(request_args, params_property_name):
    machine_input = json.loads(request_args[params_property_name])
    machine_input[STEP_DICT_NAME] = {'id': str(uuid4()), 'step_num': -1}
    request_args[params_property_name] = json.dumps(machine_input)


def handle_stepfunc_args(args):
    try:
        event_operation, request_args = args

        if event_operation == 'StartExecution':
            add_steps_dict_to_request(request_args, 'input')
        elif event_operation == 'SendTaskSuccess':
            add_steps_dict_to_request(request_args, 'output')
    except Exception:  # pylint: disable=broad-except
        pass


def patch():
    """
    Patch module.
    :return: None
    """

    wrapt.wrap_function_wrapper(
        'botocore.client',
        'BaseClient._make_api_call',
        _wrapper
    )

    # botocore no longer vendor requests in new version
    # https://github.com/boto/botocore/pull/1829
    wrapt.wrap_function_wrapper(
        'botocore.vendored.requests',
        'Session.send',
        _requests_wrapper
    )



"""
logging patcher module.
"""

from __future__ import absolute_import

import json
import os
from functools import partial

import wrapt

from ..trace import trace_factory
from ..utils import print_debug

LOGGING_FUNCTIONS = (
    'info',
    'debug',
    'error',
    'warning',
    'exception',
    'critical'
)


def _wrapper(wrapped, _instance, args, kwargs):
    """
    Wrapper for logging module.
    :param wrapped: wrapt's wrapped
    :param _instance: wrapt's instance, unused.
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """
    if not os.getenv('EPSAGON_DISABLE_LOGGING_ERRORS', '').upper() == 'TRUE':
        try:
            message = args[0] % args[1:]
            trace_factory.set_error(message)
        except Exception:  # pylint: disable=broad-except
            print_debug('Could not capture exception from log: {}'.format(
                args
            ))

    return wrapped(*args, **kwargs)


def _add_log_id(trace_log_id, msg):
    """
    adds log id to the msg
    """
    try:
        # Check if message is in json format
        json_log = json.loads(msg)
        json_log['epsagon'] = {'trace_id': trace_log_id}
        return json.dumps(json_log)
    except Exception:   # pylint: disable=broad-except
        # message is a regular string, add the ID to the beginning
        if not isinstance(msg, str):
            msg = str(msg)
        return ' '.join([trace_log_id, msg])


def _epsagon_trace_id_wrapper(msg_index, wrapped, _instance, args, kwargs):
    """
    Wrapper for logging module.
    :param msg_index: the index of the log message in args
        (since Logger.log also gets `level`)
    :param wrapped: wrapt's wrapped
    :param _instance: wrapt's instance, unused.
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """
    trace_log_id = trace_factory.get_log_id()

    if not trace_log_id:
        return wrapped(*args, **kwargs)

    try:
        message = _add_log_id(trace_log_id, args[msg_index])
    except Exception:   # pylint: disable=broad-except
        # total failure to add log id
        return wrapped(*args, **kwargs)
    args = (
        args[0:msg_index] +
        (message,) +
        args[(msg_index + 1):]
    )
    return wrapped(*args, **kwargs)


def patch():
    """
    Patch module.
    :return: None
    """
    # Automatically capture exceptions from logging
    wrapt.wrap_function_wrapper('logging', 'exception', _wrapper)
    wrapt.wrap_function_wrapper('logging', 'Logger.exception', _wrapper)

    # Instrument logging with Epsagon trace ID
    if trace_factory.is_logging_tracing_enabled():
        wrapt.wrap_function_wrapper(
            'logging',
            'Logger.log',
            partial(_epsagon_trace_id_wrapper, 1)
        )
        for log_function in LOGGING_FUNCTIONS:
            wrapt.wrap_function_wrapper(
                'logging',
                'Logger.{}'.format(log_function),
                partial(_epsagon_trace_id_wrapper, 0)
            )

    # Instrument print function is disabled
    # wrapt.wrap_function_wrapper(
    #   'builtins',
    #   'print',
    #   partial(_epsagon_trace_id_wrapper, 0)
    # )



import json
import mock
import pytest
import warnings
import epsagon.wrappers.aws_lambda
from epsagon import trace_factory
from epsagon.trace import FAILED_TO_SERIALIZE_MESSAGE
from epsagon.runners.aws_lambda import LambdaRunner, StepLambdaRunner
import epsagon.constants
from .common import get_tracer_patch_kwargs

trace_mock = mock.MagicMock()


def setup_function(func):
    trace_factory.use_single_trace = True
    trace_mock.configure_mock(**get_tracer_patch_kwargs())
    epsagon.constants.COLD_START = True


def _get_runner_event(trace_mock, runner_type=LambdaRunner):
    for args, _ in trace_mock.set_runner.call_args_list:
        event = args[0]
        if isinstance(event, runner_type):
            return event

    assert False, "No runner found"


CONTEXT_STUB = type(
    'Context',
    (object,),
    {
        'aws_request_id': 'test_request_id',
        'function_name': 'TestFunction',
        'log_stream_name': 'test_stream',
        'log_group_name': 'test_group',
        'function_version': 'test_version',
        'memory_limit_in_mb': '1024',
        'invoked_function_arn': 'arn:aws:lambda:us-east-1:123456789012:function:TestFunction'
    }
)


# aws_lambda tests
@mock.patch.object(LambdaRunner, 'set_exception')
@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.triggers.aws_lambda.LambdaTriggerFactory.factory',
    side_effect=['trigger']
)
def test_lambda_wrapper_sanity(
        trigger_factory_mock,
        _,
        set_exception_mock
):
    retval = 'success'

    @epsagon.wrappers.aws_lambda.lambda_wrapper
    def wrapped_lambda(event, context):
        return 'success'

    assert wrapped_lambda('a', CONTEXT_STUB) == 'success'
    trace_mock.prepare.assert_called()
    runner = _get_runner_event(trace_mock)

    trigger_factory_mock.assert_called()
    set_exception_mock.assert_not_called()

    trace_mock.set_timeout_handler.assert_called()

    trace_mock.send_traces.assert_called()
    trace_mock.add_exception.assert_not_called()

    assert not epsagon.constants.COLD_START
    assert runner.resource['metadata']['return_value'] == retval


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.triggers.aws_lambda.LambdaTriggerFactory.factory',
    side_effect=['trigger']
)
def test_lambda_wrapper_lambda_exception(trigger_factory_mock, _):
    @epsagon.wrappers.aws_lambda.lambda_wrapper
    def wrapped_lambda(event, context):
        raise TypeError('test')

    lambda_runner_mock = mock.MagicMock(set_exception=mock.MagicMock())
    with mock.patch(
            'epsagon.runners.aws_lambda.LambdaRunner',
            side_effect=[lambda_runner_mock]
    ):
        with pytest.raises(TypeError):
            wrapped_lambda('a', CONTEXT_STUB)

    trigger_factory_mock.assert_called()

    lambda_runner_mock.set_exception.assert_called()

    trace_mock.prepare.assert_called()
    trace_mock.add_event.assert_called()
    trace_mock.send_traces.assert_called()
    trace_mock.add_exception.assert_not_called()

    assert not epsagon.constants.COLD_START


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
def test_lambda_wrapper_lambda_exception_args(_):
    """
    Tests that when user invoking Lambda's handler manually with kwargs,
    trace won't be sent, and return value is ok.
    """

    @epsagon.wrappers.aws_lambda.lambda_wrapper
    def wrapped_lambda(event, context):
        return 'success'

    lambda_runner_mock = mock.MagicMock(set_exception=mock.MagicMock())
    with mock.patch(
            'epsagon.runners.aws_lambda.LambdaRunner',
            side_effect=[lambda_runner_mock]
    ):
        assert wrapped_lambda(event='a', context='b') == 'success'

    trace_mock.prepare.assert_called()
    trace_mock.add_event.assert_not_called()
    trace_mock.send_traces.assert_not_called()
    trace_mock.add_exception.assert_not_called()
    assert epsagon.constants.COLD_START


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.triggers.aws_lambda.LambdaTriggerFactory.factory',
    side_effect=TypeError()
)
def test_lambda_wrapper_trigger_exception(trigger_factory_mock, _):
    @epsagon.wrappers.aws_lambda.lambda_wrapper
    def wrapped_lambda(event, context):
        pass

    lambda_runner_mock = mock.MagicMock(set_exception=mock.MagicMock())
    with mock.patch(
            'epsagon.runners.aws_lambda.LambdaRunner',
            side_effect=[lambda_runner_mock]
    ):
        wrapped_lambda('a', CONTEXT_STUB)

    trigger_factory_mock.assert_called()

    lambda_runner_mock.set_exception.assert_not_called()

    trace_mock.prepare.assert_called()
    trace_mock.set_runner.assert_called()
    trace_mock.send_traces.assert_called()
    trace_mock.add_exception.assert_called()

    assert not epsagon.constants.COLD_START


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.wrappers.python_function.wrap_python_function',
    side_effect=['success']
)
def test_lambda_wrapper_none_context(wrap_python_function_wrapper, _):
    @epsagon.wrappers.aws_lambda.lambda_wrapper
    def wrapped_lambda(event, context):
        # Doesn't matter, we are mocking wrap_python_function
        return 'something'

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        lambda_runner_mock = mock.MagicMock(set_exception=mock.MagicMock())
        with mock.patch(
                'epsagon.runners.aws_lambda.LambdaRunner',
                side_effect=TypeError()
        ):
            assert wrapped_lambda('a', None) == 'success'
        assert len(w) == 1

    trace_mock.prepare.assert_called()
    wrap_python_function_wrapper.assert_called()


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.wrappers.python_function.wrap_python_function',
    side_effect=['success']
)
def test_lambda_wrapper_lambda_runner_factory_failed(
        wrap_python_function_wrapper, _):
    @epsagon.wrappers.aws_lambda.lambda_wrapper
    def wrapped_lambda(event, context):
        # Doesn't matter, we are mocking wrap_python_function
        return 'something'

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        with mock.patch(
                'epsagon.runners.aws_lambda.LambdaRunner',
                side_effect=TypeError()
        ):
            assert wrapped_lambda('a', CONTEXT_STUB) == 'success'
        assert len(w) == 1

    trace_mock.prepare.assert_called()
    wrap_python_function_wrapper.assert_called()


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
def test_lambda_wrapper_result_status_code(_):
    result = {'statusCode': 200}

    @epsagon.wrappers.aws_lambda.lambda_wrapper
    def wrapped_function(event, context):
        return result

    assert wrapped_function('a', CONTEXT_STUB) == result
    trace_mock.prepare.assert_called_once()
    runner = _get_runner_event(trace_mock)

    trace_mock.send_traces.assert_called_once()
    trace_mock.add_exception.assert_not_called()

    assert (
            runner.resource['metadata']['status_code'] == result['statusCode']
    )


# step_lambda_wrapper tests

@mock.patch.object(StepLambdaRunner, 'set_exception')
@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.triggers.aws_lambda.LambdaTriggerFactory.factory',
    side_effect=['trigger']
)
def test_step_lambda_wrapper_sanity_first_step(
        trigger_factory_mock,
        _,
        set_exception_mock
):
    retval = {'result': 'success'}

    @epsagon.wrappers.aws_lambda.step_lambda_wrapper
    def wrapped_lambda(event, context):
        return retval

    result = wrapped_lambda('a', CONTEXT_STUB)
    assert ('result', 'success',) in result.items()
    assert 'Epsagon' in result
    assert ('step_num', 0,) in result['Epsagon'].items()
    assert 'id' in result['Epsagon']

    trigger_factory_mock.assert_called()
    set_exception_mock.assert_not_called()

    trace_mock.prepare.assert_called()
    trace_mock.add_event.assert_called()
    trace_mock.set_runner.assert_called()
    trace_mock.set_timeout_handler.assert_called()
    runner = _get_runner_event(trace_mock, runner_type=StepLambdaRunner)
    trace_mock.send_traces.assert_called()
    trace_mock.add_exception.assert_not_called()

    assert not epsagon.constants.COLD_START
    assert runner.resource['metadata']['return_value'] == retval


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.triggers.aws_lambda.LambdaTriggerFactory.factory',
    side_effect=['trigger']
)
def test_step_lambda_wrapper_sanity_not_first_step(trigger_factory_mock,
                                                   _):
    @epsagon.wrappers.aws_lambda.step_lambda_wrapper
    def wrapped_lambda(event, context):
        return {'result': 'success'}

    lambda_runner_mock = mock.MagicMock(set_exception=mock.MagicMock())
    with mock.patch(
            'epsagon.runners.aws_lambda.StepLambdaRunner',
            side_effect=[lambda_runner_mock]
    ):
        result = wrapped_lambda(
            {'a': 'a', 'Epsagon': {'step_num': 1, 'id': 1}},
            CONTEXT_STUB
        )
        assert ('result', 'success',) in result.items()
        assert 'Epsagon' in result
        assert ('step_num', 2,) in result['Epsagon'].items()
        assert 'id' in result['Epsagon']

    trigger_factory_mock.assert_called()
    lambda_runner_mock.set_exception.assert_not_called()

    trace_mock.prepare.assert_called()
    trace_mock.add_event.assert_called()
    trace_mock.send_traces.assert_called()
    trace_mock.add_exception.assert_not_called()

    assert not epsagon.constants.COLD_START


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.triggers.aws_lambda.LambdaTriggerFactory.factory',
    side_effect=['trigger']
)
def test_step_lambda_wrapper_sanity_not_first_step_unqiue_path(trigger_factory_mock,
                                                   _):
    unique_path = 'data'
    @epsagon.wrappers.aws_lambda.step_lambda_wrapper
    def wrapped_lambda(event, context):
        return {'result': 'success', unique_path: {'test':' test'}}

    lambda_runner_mock = mock.MagicMock(set_exception=mock.MagicMock())
    with mock.patch(
            'epsagon.runners.aws_lambda.StepLambdaRunner',
            side_effect=[lambda_runner_mock]
    ):
        result = wrapped_lambda(
            {'a': 'a', unique_path: {'Epsagon': {'step_num': 1, 'id': 1}}},
            CONTEXT_STUB
        )
        assert ('result', 'success',) in result.items()
        assert 'Epsagon' in result[unique_path]
        assert ('step_num', 2,) in result[unique_path]['Epsagon'].items()
        assert 'id' in result[unique_path]['Epsagon']

    trigger_factory_mock.assert_called()
    lambda_runner_mock.set_exception.assert_not_called()

    trace_mock.prepare.assert_called()
    trace_mock.add_event.assert_called()
    trace_mock.send_traces.assert_called()
    trace_mock.add_exception.assert_not_called()

    assert not epsagon.constants.COLD_START


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.triggers.aws_lambda.LambdaTriggerFactory.factory',
    side_effect=['trigger']
)
def test_step_lambda_wrapper_wrapped_function_doesnt_return_object(
        trigger_factory_mock, _):
    @epsagon.wrappers.aws_lambda.step_lambda_wrapper
    def wrapped_lambda(event, context):
        return 'success'

    lambda_runner_mock = mock.MagicMock(set_exception=mock.MagicMock())
    with mock.patch(
            'epsagon.runners.aws_lambda.StepLambdaRunner',
            side_effect=[lambda_runner_mock]
    ):
        assert wrapped_lambda(
            {'a': 'a', 'Epsagon': {'step_num': 1, 'id': 1}},
            CONTEXT_STUB
        ) == 'success'

    trigger_factory_mock.assert_called()
    lambda_runner_mock.set_exception.assert_not_called()

    trace_mock.prepare.assert_called()
    trace_mock.add_event.assert_called()
    trace_mock.send_traces.assert_called()
    trace_mock.add_exception.assert_not_called()

    assert not epsagon.constants.COLD_START


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.triggers.aws_lambda.LambdaTriggerFactory.factory',
    side_effect=['trigger']
)
def test_step_lambda_wrapper_lambda_exception(trigger_factory_mock, _):
    @epsagon.wrappers.aws_lambda.step_lambda_wrapper
    def wrapped_lambda(event, context):
        raise TypeError('test')

    lambda_runner_mock = mock.MagicMock(set_exception=mock.MagicMock())
    with mock.patch(
            'epsagon.runners.aws_lambda.StepLambdaRunner',
            side_effect=[lambda_runner_mock]
    ):
        with pytest.raises(TypeError):
            wrapped_lambda('a', CONTEXT_STUB)

    trigger_factory_mock.assert_called()

    lambda_runner_mock.set_exception.assert_called()

    trace_mock.prepare.assert_called()
    trace_mock.add_event.assert_called()
    trace_mock.send_traces.assert_called()
    trace_mock.add_exception.assert_not_called()

    assert not epsagon.constants.COLD_START


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.triggers.aws_lambda.LambdaTriggerFactory.factory',
    side_effect=TypeError()
)
def test_step_lambda_wrapper_trigger_exception(trigger_factory_mock,
                                               _):
    @epsagon.wrappers.aws_lambda.step_lambda_wrapper
    def wrapped_lambda(event, context):
        pass

    lambda_runner_mock = mock.MagicMock(set_exception=mock.MagicMock())
    with mock.patch(
            'epsagon.runners.aws_lambda.StepLambdaRunner',
            side_effect=[lambda_runner_mock]
    ):
        wrapped_lambda('a', CONTEXT_STUB)

    trigger_factory_mock.assert_called()

    lambda_runner_mock.set_exception.assert_not_called()

    trace_mock.prepare.assert_called()
    trace_mock.set_runner.assert_called()
    trace_mock.send_traces.assert_called()
    trace_mock.add_exception.assert_called()

    assert not epsagon.constants.COLD_START


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.wrappers.python_function.wrap_python_function',
    side_effect=['success']
)
def test_step_lambda_wrapper_none_context(wrap_python_function_wrapper,
                                          _):
    @epsagon.wrappers.aws_lambda.step_lambda_wrapper
    def wrapped_lambda(event, context):
        # Doesn't matter, we are mocking wrap_python_function
        return 'something'

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        lambda_runner_mock = mock.MagicMock(set_exception=mock.MagicMock())
        with mock.patch(
                'epsagon.runners.aws_lambda.StepLambdaRunner',
                side_effect=TypeError()
        ):
            assert wrapped_lambda('a', None) == 'success'
        assert len(w) == 1

    trace_mock.prepare.assert_called()
    wrap_python_function_wrapper.assert_called()


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.wrappers.python_function.wrap_python_function',
    side_effect=['success']
)
def test_step_lambda_wrapper_lambda_runner_factory_failed(
        wrap_python_function_wrapper, _):
    @epsagon.wrappers.aws_lambda.step_lambda_wrapper
    def wrapped_lambda(event, context):
        # Doesn't matter, we are mocking wrap_python_function
        return 'something'

    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter('always')
        with mock.patch(
                'epsagon.runners.aws_lambda.StepLambdaRunner',
                side_effect=TypeError()
        ):
            assert wrapped_lambda('a', CONTEXT_STUB) == 'success'
        assert len(w) == 1

    trace_mock.prepare.assert_called()
    wrap_python_function_wrapper.assert_called()


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
def test_step_lambda_wrapper_result_status_code(_):
    result = {'statusCode': 200}

    @epsagon.wrappers.aws_lambda.step_lambda_wrapper
    def wrapped_function(event, context):
        return result

    assert wrapped_function('a', CONTEXT_STUB) == result
    trace_mock.prepare.assert_called_once()
    runner = _get_runner_event(trace_mock, runner_type=StepLambdaRunner)

    trace_mock.send_traces.assert_called_once()
    trace_mock.add_exception.assert_not_called()

    assert (
            runner.resource['metadata']['status_code'] ==
            result['statusCode']
    )


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
def test_lambda_wrapper_with_alias_arn(
        _,
):
    @epsagon.wrappers.aws_lambda.lambda_wrapper
    def wrapped_lambda(event, context):
        return 'success'

    orig_arn = CONTEXT_STUB.invoked_function_arn
    CONTEXT_STUB.invoked_function_arn = (
        'arn:aws:lambda:us-east-1:123456789012:function:TestFunction:test_alias'
    )
    assert wrapped_lambda('a', CONTEXT_STUB) == 'success'
    CONTEXT_STUB.invoked_function_arn = orig_arn
    trace_mock.prepare.assert_called()
    runner = _get_runner_event(trace_mock)

    assert runner.resource['metadata']['function_alias'] == 'test_alias'


@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
def test_lambda_wrapper_single_thread(_):
    retval = 'success'

    @epsagon.wrappers.aws_lambda.lambda_wrapper
    def wrapped_lambda(event, context):
        return retval

    assert wrapped_lambda('a', CONTEXT_STUB) == retval
    assert trace_factory.use_single_trace


def test_lambda_wrapper_avoid_multi_wrap():
    @epsagon.wrappers.aws_lambda.lambda_wrapper
    def wrapped_lambda(event, context):
        return 'success'

    new_wrapped = epsagon.wrappers.aws_lambda.lambda_wrapper(wrapped_lambda)
    assert getattr(wrapped_lambda, '__instrumented__', False) == True
    assert new_wrapped == wrapped_lambda


@mock.patch.object(LambdaRunner, 'set_exception')
@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.triggers.aws_lambda.LambdaTriggerFactory.factory',
    side_effect=['trigger']
)
def test_propagate_lambda_id_to_dict_sanity(
        trigger_factory_mock,
        _,
        set_exception_mock
):
    retval = {
        'hello': 2,
        epsagon.constants.EPSAGON_EVENT_ID_KEY: 'test_request_id'
    }

    trace_mock.propagate_lambda_id = True

    @epsagon.wrappers.aws_lambda.lambda_wrapper
    def wrapped_lambda(_event, _context):
        return {'hello': 2}

    assert wrapped_lambda('a', CONTEXT_STUB) == retval
    trace_mock.propagate_lambda_id = False

    trace_mock.prepare.assert_called()
    runner = _get_runner_event(trace_mock)

    trigger_factory_mock.assert_called()
    set_exception_mock.assert_not_called()

    trace_mock.set_timeout_handler.assert_called()

    trace_mock.send_traces.assert_called()
    trace_mock.add_exception.assert_not_called()
    assert not epsagon.constants.COLD_START
    assert runner.resource['metadata']['return_value'] == retval


@mock.patch.object(LambdaRunner, 'set_exception')
@mock.patch(
    'epsagon.trace.trace_factory.get_or_create_trace',
    side_effect=lambda: trace_mock
)
@mock.patch(
    'epsagon.triggers.aws_lambda.LambdaTriggerFactory.factory',
    side_effect=['trigger']
)
def test_skip_propagate_lambda_id_to_non_dict_sanity(
        trigger_factory_mock,
        _,
        set_exception_mock
):
    retval = 'hey'

    trace_mock.propagate_lambda_id = True

    @epsagon.wrappers.aws_lambda.lambda_wrapper
    def wrapped_lambda(_event, _context):
        return 'hey'

    assert wrapped_lambda('a', CONTEXT_STUB) == retval
    trace_mock.propagate_lambda_id = False

    trace_mock.prepare.assert_called()
    runner = _get_runner_event(trace_mock)

    trigger_factory_mock.assert_called()
    set_exception_mock.assert_not_called()

    trace_mock.set_timeout_handler.assert_called()

    trace_mock.send_traces.assert_called()
    trace_mock.add_exception.assert_not_called()
    assert not epsagon.constants.COLD_START
    assert runner.resource['metadata']['return_value'] == retval



"""
Redis patcher module.
"""

from __future__ import absolute_import
import copy
import wrapt
from epsagon.modules.general_wrapper import wrapper
from ..events.redis import RedisSingleEventFactory, RedisMultiEventFactory


def _single_wrapper(wrapped, instance, args, kwargs):
    """
    Single execution wrapper for Redis instrumentation.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """
    return wrapper(RedisSingleEventFactory, wrapped, instance, args, kwargs)


def _multi_wrapper(wrapped, instance, args, kwargs):
    """
    Multi-execution wrapper for Redis instrumentation.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """
    RedisMultiEventFactory.LAST_STACK = copy.deepcopy(instance.command_stack)
    return wrapper(RedisMultiEventFactory, wrapped, instance, args, kwargs)


def patch():
    """
    patch module.
    :return: None
    """
    wrapt.wrap_function_wrapper(
        'redis',
        'Redis.execute_command',
        _single_wrapper
    )
    wrapt.wrap_function_wrapper(
        'redis.client',
        'Pipeline.immediate_execute_command',
        _single_wrapper
    )
    wrapt.wrap_function_wrapper(
        'redis.client',
        'Pipeline.execute',
        _multi_wrapper
    )



"""
psycopg2 patcher module
"""
from __future__ import absolute_import

import wrapt
from .db_wrapper import connect_wrapper


#pylint: disable=W0613
def _register_type_wrapper(wrapped, instance, args, kwargs):
    """
    register_type wrapper for psycopg2 instrumentation
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """

    def _extract_arguments(obj, scope=None):
        return obj, scope

    obj, scope = _extract_arguments(*args, **kwargs)

    if scope is not None:
        if isinstance(scope, wrapt.ObjectProxy):
            scope = scope.__wrapped__
        return wrapped(obj, scope)

    return wrapped(obj)


# pylint: disable=abstract-method
class AdapterWrapper(wrapt.ObjectProxy):
    """
    a wrapper for an adapter, to strip the connection out of the objectProxy
    before calling prepare
    """

    def prepare(self, *args, **kwargs):
        """
        Prepare wrapper.
        :param args:
        :param kwargs:
        :return:
        """
        if not args:
            return self.__wrapped__.prepare(*args, **kwargs)

        connection = args[0]
        if isinstance(connection, wrapt.ObjectProxy):
            connection = connection.__wrapped__

        return self.__wrapped__.prepare(connection, *args[1:], **kwargs)


def _adapt_wrapper(wrapped, instance, args, kwargs):
    """
    adapt wrapper for psycopg2 instrumentation
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """

    adapter = wrapped(*args, **kwargs)
    return AdapterWrapper(adapter) if hasattr(adapter, 'prepare') else adapter


def _patch_unwrappers():
    """
    patches the functions that do not accept our ObjectProxy to strip the proxy
    before calling the function
    :return:
    """

    wrapt.wrap_function_wrapper(
        'psycopg2.extensions',
        'register_type',
        _register_type_wrapper
    )

    wrapt.wrap_function_wrapper(
        'psycopg2._psycopg',
        'register_type',
        _register_type_wrapper
    )

    wrapt.wrap_function_wrapper(
        'psycopg2._json',
        'register_type',
        _register_type_wrapper
    )

    wrapt.wrap_function_wrapper(
        'psycopg2.extensions',
        'adapt',
        _adapt_wrapper
    )


def patch():
    """
    patch module.
    :return: None
    """

    wrapt.wrap_function_wrapper(
        'psycopg2',
        'connect',
        connect_wrapper
    )
    _patch_unwrappers()



"""
urllib3 events module.
"""

from __future__ import absolute_import

try:
    from urllib.parse import urlparse, urlunparse
except ImportError:
    from urlparse import urlparse, urlunparse
import traceback
from uuid import uuid4

from epsagon.utils import add_data_if_needed
from ..trace import trace_factory
from ..event import BaseEvent
from ..http_filters import (
    is_blacklisted_url,
    is_payload_collection_blacklisted
)
from ..utils import update_http_headers, normalize_http_url
from ..constants import HTTP_ERR_CODE, EPSAGON_MARKER


class Urllib3Event(BaseEvent):
    """
    Represents base requests event.
    """

    ORIGIN = 'urllib3'
    RESOURCE_TYPE = 'http'

    # pylint: disable=W0613
    def __init__(self, wrapped, instance, args, kwargs, start_time, response,
                 exception):
        """
        Initialize.
        :param wrapped: wrapt's wrapped
        :param instance: wrapt's instance
        :param args: wrapt's args
        :param kwargs: wrapt's kwargs
        :param start_time: Start timestamp (epoch)
        :param response: response data
        :param exception: Exception (if happened)
        """
        super(Urllib3Event, self).__init__(start_time)

        self.event_id = 'urllib3-{}'.format(str(uuid4()))

        method, url = args
        body = kwargs.get('body')
        headers = kwargs.get('headers')
        if headers:
            # Make sure trace ID is present in case headers will be removed.
            epsagon_trace_id = headers.get('epsagon-trace-id')
            if epsagon_trace_id:
                self.resource['metadata']['http_trace_id'] = epsagon_trace_id

        parsed_url = urlparse(url)
        # Omitting ports (`:80'/':443') for the host URL.
        host_url = parsed_url.netloc.split(':')[0]
        full_url = urlunparse((
            parsed_url.scheme,
            host_url,
            parsed_url.path,
            parsed_url.params,
            parsed_url.query,
            parsed_url.fragment,
        ))

        self.resource['name'] = normalize_http_url(url)
        self.resource['operation'] = method
        self.resource['metadata']['url'] = full_url

        if not is_payload_collection_blacklisted(full_url):
            add_data_if_needed(
                self.resource['metadata'],
                'request_headers',
                dict(headers)
            )

            add_data_if_needed(
                self.resource['metadata'],
                'request_body',
                body
            )

        if response is not None:
            self.update_response(response)

        if exception is not None:
            self.set_exception(exception, traceback.format_exc())

    def update_response(self, response):
        """
        Adds response data to event.
        :param response: Response from botocore
        :return: None
        """
        self.resource['metadata']['status_code'] = response.status
        headers = dict(response.getheaders())
        self.resource = update_http_headers(
            self.resource,
            headers
        )

        self.resource['metadata']['response_body'] = None
        full_url = self.resource['metadata']['url']

        if not is_payload_collection_blacklisted(full_url):
            add_data_if_needed(
                self.resource['metadata'],
                'response_headers',
                headers
            )

        # Detect errors based on status code
        if response.status >= HTTP_ERR_CODE:
            self.set_error()


class Urllib3EventFactory(object):
    """
    Factory class, generates urllib3 event.
    """

    @staticmethod
    def create_event(wrapped, instance, args, kwargs, start_time, response,
                     exception):
        """
        Create an event according to the given api_name.
        """
        # Ignore requests events since they are being instrumented.
        is_requests = (
            str(kwargs.get('headers', {}).get('User-Agent', '')).startswith(
                'python-requests'
            )
        )
        if is_requests or getattr(instance, EPSAGON_MARKER, False):
            return

        path = args[1] if len(args) > 1 else kwargs.get('url', '')
        port_part = ':' + str(instance.port) if instance.port else ''
        host_url = '{}://{}'.format(instance.scheme, instance.host)
        url = '{}{}{}'.format(
            host_url, port_part, path
        )
        args = (args[0] if args else kwargs.get('method', 'UNKNOWN METHOD'),
                url)

        # Detect if URL is blacklisted, and ignore.
        if is_blacklisted_url(host_url):
            return

        event = Urllib3Event(
            wrapped,
            instance,
            args,
            kwargs,
            start_time,
            response,
            exception
        )

        trace_factory.add_event(event)



"""
Wrapper for Django.
"""
from __future__ import absolute_import

import time
import traceback
import warnings
import epsagon
import epsagon.trace
import epsagon.triggers.http
import epsagon.runners.django

from epsagon.common import EpsagonWarning
from epsagon.utils import (
    collect_container_metadata,
    get_traceback_data_from_exception,
)
from ..http_filters import ignore_request


class DjangoMiddleware(object):
    """
    Represents a Django Middleware for Epsagon instrumentation.
    """

    def __init__(self, get_response):
        self.get_response = get_response

        epsagon.trace.trace_factory.switch_to_multiple_traces()

    # pylint: disable=no-self-use
    def process_exception(self, request, process_exception):
        """
        Processes and appends a given exception to the current trace
        """
        if not process_exception:
            return

        if (
                not hasattr(request, 'epsagon_trace') or
                not request.epsagon_trace.runner
        ):
            return

        traceback_data = get_traceback_data_from_exception(process_exception)

        request.epsagon_trace.runner.set_exception(
            process_exception,
            traceback_data,
            False
        )

    def __call__(self, request):
        # Link epsagon to the request object for easy-access to epsagon lib
        request.epsagon = epsagon

        if epsagon.http_filters.is_ignored_endpoint(request.path):
            return self.get_response(request)

        request_middleware = DjangoRequestMiddleware(request)
        request_middleware.before_request()

        response = self.get_response(request)

        request_middleware.after_request(response)
        return response


class DjangoRequestMiddleware(object):
    """
    Django middleware for a single request
    """

    def __init__(self, request):
        self.request = request
        self.runner = None
        self.ignored_request = False

    def before_request(self):
        """
        Runs before process of response.
        """
        trace = epsagon.trace.trace_factory.get_or_create_trace()
        trace.prepare()

        # Ignoring non relevant content types.
        self.ignored_request = ignore_request('', self.request.path.lower())

        if self.ignored_request:
            return

        # Create a Django runner with current request.
        try:
            self.runner = epsagon.runners.django.DjangoRunner(
                time.time(),
                self.request
            )
            trace.set_runner(self.runner)

            # Collect metadata in case this is a container.
            collect_container_metadata(self.runner.resource['metadata'])

        # pylint: disable=W0703
        except Exception as exception:
            # Regress to python runner.
            warnings.warn('Could not extract request', EpsagonWarning)
            epsagon.trace.trace_factory.add_exception(
                exception,
                traceback.format_exc()
            )

        # Extract HTTP trigger data.
        try:
            trigger = epsagon.triggers.http.HTTPTriggerFactory.factory(
                time.time(),
                self.request
            )
            if trigger:
                epsagon.trace.trace_factory.add_event(trigger)
        # pylint: disable=W0703
        except Exception as exception:
            epsagon.trace.trace_factory.add_exception(
                exception,
                traceback.format_exc(),
            )

    def after_request(self, response):
        """
        Runs after process of response.
        """
        if self.ignored_request:
            return

        # Ignoring non relevant content types.
        if ignore_request(response.get('Content-Type', '').lower(), ''):
            self.ignored_request = True
            return

        # Safety in case we run on an old Django version
        if not self.runner:
            return

        self.runner.update_response(response)
        epsagon.trace.trace_factory.send_traces()



"""
Base Event class
"""

from __future__ import absolute_import
import time
import uuid
from .common import ErrorCode


class BaseEvent(object):
    """
    Represents base trace's event
    """

    ORIGIN = 'base'
    RESOURCE_TYPE = 'base'

    def __init__(self, start_time):
        """
        Initialize.
        :param start_time: event's start time (epoch)
        """

        self.start_time = start_time
        self.event_id = ''
        self.origin = self.ORIGIN
        self.duration = 0.0
        self.error_code = ErrorCode.OK
        self.exception = {}
        self.terminated = False

        self.resource = {
            'type': self.RESOURCE_TYPE,
            'name': '',
            'operation': '',
            'metadata': {},
        }

        if self.origin == 'runner':
            self.resource['metadata']['trace_id'] = str(uuid.uuid4())

    @staticmethod
    def load_from_dict(event_data):
        """
        Load Event object from dict.
        :param event_data: dict
        :return: Event
        """

        event = BaseEvent(event_data['start_time'])
        event.event_id = event_data['id']
        event.origin = event_data['origin']
        event.duration = event_data['duration']
        event.error_code = event_data['error_code']
        event.resource = event_data['resource']
        if event.error_code == ErrorCode.EXCEPTION:
            event.exception = event_data['exception']

        return event

    def to_dict(self):
        """
        Converts Event to dict.
        :return: dict
        """

        self_as_dict = {
            'id': self.event_id,
            'start_time': self.start_time,
            'duration': self.duration,
            'origin': self.origin,
            'error_code': self.error_code,
            'resource': self.resource,
        }

        if self.error_code == ErrorCode.EXCEPTION:
            self_as_dict['exception'] = self.exception

        return self_as_dict

    def terminate(self):
        """
        Sets duration time.
        :return: None
        """
        if not self.terminated:
            self.duration = time.time() - self.start_time
            self.terminated = True

    def set_error(self):
        """
        Sets general error.
        :return: None
        """
        if self.error_code != ErrorCode.EXCEPTION:
            self.error_code = ErrorCode.ERROR

    def set_exception(self, exception, traceback_data, handled=True):
        """
        Sets exception data on event.
        :param exception: Exception object
        :param traceback_data: traceback string
        :param handled: False if the exception was raised from the wrapped
            function
        """

        self.error_code = ErrorCode.EXCEPTION
        self.exception['type'] = type(exception).__name__
        self.exception['message'] = str(exception)
        self.exception['traceback'] = traceback_data
        self.exception['time'] = time.time()
        self.exception.setdefault('additional_data', {})['handled'] = handled



"""
Utils for web frameworks request filters.
"""

from six.moves import urllib
from epsagon.trace import trace_factory
from epsagon.constants import IGNORED_ENDPOINTS

# Ignored content types for web frameworks.
IGNORED_CONTENT_TYPES = [
    'image',
    'audio',
    'video',
    'font',
    'zip',
    'css',
]
IGNORED_FILE_TYPES = [
    '.js',
    '.jsx',
    '.woff',
    '.woff2',
    '.ttf',
    '.eot',
    '.ico',
]

# Method to URL dict.
BLACKLIST_URLS = {
    str.endswith: [
        'epsagon.com',
        '.amazonaws.com',
    ],
    str.__contains__: [
        'accounts.google.com',
        'documents.azure.com',
        '169.254.170.2'  # AWS Task Metadata Endpoint
    ],
}
WHITELIST_URL = {
    str.__contains__: [
        '.execute-api.',
        '.elb.amazonaws.com',
        '.appsync-api.',
    ],
}


def is_blacklisted_url(url):
    """
    Return whether the URL blacklisted or not.
    Using BLACKLIST_URLS methods against the URLs.
    :param url: url string
    :return: True if URL is blacklisted, else False
    """

    url = urllib.parse.urlparse(url).netloc
    for method in WHITELIST_URL:
        for whitelist_url in WHITELIST_URL[method]:
            if method(url, whitelist_url):
                return False

    for method in BLACKLIST_URLS:
        for blacklist_url in BLACKLIST_URLS[method]:
            if method(url, blacklist_url):
                return True

    return False


def is_payload_collection_blacklisted(url):
    """
    Return whether the payload should be collected according to the blacklisted
    urls list in the Trace.
    :param url: url string
    :return:  True if URL is blacklisted, else False
    """
    url = urllib.parse.urlparse(url).netloc
    trace_blacklist_urls = trace_factory.get_trace().url_patterns_to_ignore
    return any(blacklist_url in url for blacklist_url in trace_blacklist_urls)


def ignore_request(content, path):
    """
    Return true if HTTP request in web frameworks should be omitted.
    :param content: accept mimetype header
    :param path: request path
    :return: Bool
    """

    return (
        any([x in content for x in IGNORED_CONTENT_TYPES]) or
        any([path.endswith(x) for x in IGNORED_FILE_TYPES])
    )


def add_ignored_endpoints(endpoints):
    """
    add endpoints to the list of ignored ones..
    :param endpoints: list of endpoints or None
    :return: None
    """
    if endpoints:
        IGNORED_ENDPOINTS.extend(endpoints)


def is_ignored_endpoint(endpoint):
    """
    return true if endpoint should be ignored.
    :param endpoint: endpoint path
    :return: Bool
    """
    return endpoint in IGNORED_ENDPOINTS



"""
Wrapper for DB modules
"""
from __future__ import absolute_import
import wrapt
import epsagon.modules.general_wrapper
from ..events.dbapi import DBAPIEventFactory


# pylint: disable=abstract-method
class CursorWrapper(wrapt.ObjectProxy):
    """
    A dbapi cursor wrapper for tracing
    """

    def __init__(self, cursor, connection_wrapper):
        super(CursorWrapper, self).__init__(cursor)
        self._self_connection = connection_wrapper

    @property
    def connection_wrapper(self):
        """
        A property that holds that connection_wrapper.
        :return: the connection wrapper.
        """
        return self._self_connection

    # NOTE: tracing other API calls currently not supported
    # (as 'executemany' and 'callproc')
    def execute(self, *args, **kwargs):
        """
        Execute the query.
        :param args: args.
        :param kwargs: kwargs.
        """
        epsagon.modules.general_wrapper.wrapper(
            DBAPIEventFactory,
            self.__wrapped__.execute,
            self,
            args,
            kwargs,
        )

    def __enter__(self):
        # raise appropriate error if api not supported (should reach the user)
        self.__wrapped__.__enter__  # pylint: disable=W0104

        return self


class ConnectionWrapper(wrapt.ObjectProxy):
    """
    A dbapi connection wrapper for tracing.
    """
    def __init__(self, connection, args, kwargs):
        super(ConnectionWrapper, self).__init__(connection)
        self._self_args = args
        self._self_kwargs = kwargs

    def cursor(self, *args, **kwargs):
        """
        Return cursor wrapper.
        :param args: args.
        :param kwargs: kwargs.
        :return: Cursorwrapper.
        """
        cursor = self.__wrapped__.cursor(*args, **kwargs)
        return CursorWrapper(cursor, self)

    @property
    def extract_hostname(self):
        """
        A property that extract the host name
        :return: the host name
        """
        return self._self_kwargs.get('host', 'local')

    @property
    def extract_dbname(self):
        """
        A property that extract the db name
        :return: the db name
        """
        return self._self_kwargs.get(
            'db',
            self._self_kwargs.get(
                'database',
                ''
            )
        )


#pylint: disable=W0613
def connect_wrapper(wrapped, instance, args, kwargs):
    """
    connect wrapper for psycopg2 instrumentation
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """
    connection = wrapped(*args, **kwargs)
    return ConnectionWrapper(connection, args, kwargs)



"""
requests patcher module.
"""

from __future__ import absolute_import
import wrapt
from epsagon.modules.general_wrapper import wrapper
from ..events.urllib import UrllibEventFactory


def _wrapper(wrapped, instance, args, kwargs):
    """
    General wrapper for requests instrumentation.
    :param wrapped: wrapt's wrapped
    :param instance: wrapt's instance
    :param args: wrapt's args
    :param kwargs: wrapt's kwargs
    :return: None
    """
    return wrapper(UrllibEventFactory, wrapped, instance, args, kwargs)


def patch():
    """
    Patch module.
    :return: None
    """

    try:
        wrapt.wrap_function_wrapper(
            'urllib.request',
            'OpenerDirector._open',
            _wrapper
        )
    except Exception:  # pylint: disable=broad-except
        # Can happen in different Python versions.
        pass
